import dataclasses
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import cast

import numpy

# this sentinel is used to omit certain dataclass fields when the dataclass is converted to a dict
SENTINEL = cast(None, "SENTINEL TO SKIP DATACLASS FIELDS WHEN CONVERTING TO DICT")


@dataclasses.dataclass
class Attribution:
    """Output dataclass for an attribution in a prediction.

    Parameters
    ----------
    text
        Text belonging to the attribution
    start
        Start char id
    end
        End char id
    field
        Field name of the input to which the attribution belongs
    attribution
        Numeric value quantifying the attribution of 'text' to the prediction
    """

    text: str
    start: int
    end: int
    field: str
    attribution: float


@dataclasses.dataclass
class Token:
    """Output dataclass for a token in a prediction.

    Parameters
    ----------
    text
        Text of the token
    start
        Start char id
    end
        End char id
    field
        Field name of the input to which the token belongs
    """

    text: str
    start: int
    end: int
    field: str


@dataclasses.dataclass
class Entity:
    """Output dataclass for a NER entity in a prediction.

    Parameters
    ----------
    label
        Label of the entity
    start_token
        Start token id
    end_token
        End token id
    start
        Start char id
    end
        End char id
    """

    label: str
    start_token: int
    end_token: int
    start: Optional[int] = SENTINEL
    end: Optional[int] = SENTINEL


@dataclasses.dataclass
class TaskPrediction:
    """Base class for the TaskOutput classes.

    Each head should implement a proper task prediction class that defines its prediction output.
    You can use the SENTINEL as default value if you want to omit certain fields when converting to a dict.
    """

    @staticmethod
    def _dict_factory(key_value_pairs: List[Tuple]) -> Dict:
        """input is a list of (key, value) pairs"""
        return dict(
            [(key, value) for key, value in key_value_pairs if value is not SENTINEL]
        )

    def as_dict(self) -> Dict:
        return dataclasses.asdict(self, dict_factory=self._dict_factory)


@dataclasses.dataclass
class TextClassificationPrediction(TaskPrediction):
    """Output dataclass for the `TextClassification` head

    Parameters
    ----------
    labels
        Ordered list of predictions, from the label with the highest to the label with the lowest probability.
    probabilities
        Ordered list of probabilities, from highest to lowest probability.
    attributions
        Attribution of each token to the prediction.
    tokens
        Tokens of the tokenized input
    """

    labels: List[str]
    probabilities: List[float]
    attributions: Optional[List[Attribution]] = SENTINEL
    tokens: Optional[List[Token]] = SENTINEL


@dataclasses.dataclass
class DocumentClassificationPrediction(TaskPrediction):
    """Output dataclass for the `DocumentClassification` head

    Parameters
    ----------
    labels
        Ordered list of predictions, from the label with the highest to the label with the lowest probability.
    probabilities
        Ordered list of probabilities, from highest to lowest probability.
    attributions
        Attribution of each token to the prediction.
    tokens
        Tokens of the tokenized input
    """

    labels: List[str]
    probabilities: List[float]
    attributions: Optional[List[List[Attribution]]] = SENTINEL
    tokens: Optional[List[List[Token]]] = SENTINEL


@dataclasses.dataclass
class RecordClassificationPrediction(TaskPrediction):
    """Output dataclass for the `RecordClassification` head

    Parameters
    ----------
    labels
        Ordered list of predictions, from the label with the highest to the label with the lowest probability.
    probabilities
        Ordered list of probabilities, from highest to lowest probability.
    attributions
        Attribution of each token to the prediction.
    tokens
        Tokens of the tokenized input
    """

    labels: List[str]
    probabilities: List[float]
    attributions: Optional[List[List[Attribution]]] = SENTINEL
    tokens: Optional[List[List[Token]]] = SENTINEL


@dataclasses.dataclass
class RecordPairClassificationPrediction(TaskPrediction):
    """Output dataclass for the `RecordPairClassification` head

    Parameters
    ----------
    labels
        Ordered list of predictions, from the label with the highest to the label with the lowest probability.
    probabilities
        Ordered list of probabilities, from highest to lowest probability.
    attributions
        Attribution of each record field to the prediction. The calculated attributions only make sense
        for a duplicate/not_duplicate binary classification task of the two records.
    tokens
        Tokens of the tokenized input
    """

    labels: List[str]
    probabilities: List[float]
    attributions: Optional[List[Attribution]] = SENTINEL
    tokens: Optional[List[List[Token]]] = SENTINEL


@dataclasses.dataclass
class RelationClassificationPrediction(TaskPrediction):
    """Output dataclass for the `RelationClassification` head

    Parameters
    ----------
    labels
        Ordered list of predictions, from the label with the highest to the label with the lowest probability.
    probabilities
        Ordered list of probabilities, from highest to lowest probability.
    tokens
        Tokens of the tokenized input
    """

    labels: List[str]
    probabilities: List[float]
    tokens: Optional[List[Token]] = SENTINEL


@dataclasses.dataclass
class TokenClassificationPrediction(TaskPrediction):
    """Output dataclass for the `TokenClassification` head

    Parameters
    ----------
    tags
        List of lists of NER tags, ordered by score.
        The list of NER tags with the highest score comes first.
    entities
        List of list of entities, ordered by score.
        The list of entities with the highest score comes first.
    scores
        Ordered list of scores for each list of NER tags (highest to lowest).
    tokens
        Tokens of the tokenized input.
    """

    tags: List[List[str]]
    entities: List[List[Entity]]
    scores: List[float]
    tokens: Optional[List[Token]] = SENTINEL


@dataclasses.dataclass
class LanguageModellingPrediction(TaskPrediction):
    """Output dataclass for the `LanguageModelling` head"""

    lm_embeddings: numpy.array
    mask: numpy.array
    # Is only included if batch size == 1
    loss: Optional[float] = SENTINEL
    tokens: Optional[List[Token]] = SENTINEL



import pytest

from biome.text import Pipeline
from biome.text.modules.heads.task_prediction import TextClassificationPrediction
from biome.text.pipeline import PredictionError


@pytest.fixture
def pipeline() -> Pipeline:
    return Pipeline.from_config(
        {
            "name": "test_predict",
            "head": {"type": "TextClassification", "labels": ["a"]},
        }
    )


def test_raise_Prediction_error(pipeline):
    with pytest.raises(PredictionError):
        pipeline.predict("")

    with pytest.raises(PredictionError):
        pipeline.predict(batch=[{"text": ""}, {"text": ""}])


def test_batch_parameter_gets_ignored(pipeline):
    prediction = pipeline.predict("testtt", batch=[{"text": "test"}], add_tokens=True)
    assert prediction["tokens"][0]["text"] == "testtt"

    prediction = pipeline.predict(
        text="testtt", batch=[{"text": "test"}], add_tokens=True
    )
    assert prediction["tokens"][0]["text"] == "testtt"


def test_map_args_kwargs_to_input():
    class MockPipeline:
        def __init__(self, inputs):
            self._inputs = inputs

        @property
        def inputs(self):
            return self._inputs

    assert Pipeline._map_args_kwargs_to_input(MockPipeline(["text"]), "test") == {
        "text": "test"
    }
    assert Pipeline._map_args_kwargs_to_input(MockPipeline(["text"]), text="test") == {
        "text": "test"
    }
    assert Pipeline._map_args_kwargs_to_input(
        MockPipeline(["text", "text2"]), "test", text2="test2"
    ) == {"text": "test", "text2": "test2"}


def test_return_single_or_list(pipeline, monkeypatch):
    def mock_predict(batch, prediction_config):
        return [
            TextClassificationPrediction(labels=["a"], probabilities=[1])
            if i % 2 == 0
            else None
            for i, _ in enumerate(batch)
        ]

    monkeypatch.setattr(pipeline._model, "predict", mock_predict)

    assert isinstance(pipeline.predict("test"), dict)

    batch_prediction = pipeline.predict(batch=[{"text": "test"}])
    assert isinstance(batch_prediction, list) and len(batch_prediction) == 1
    assert isinstance(batch_prediction[0], dict)

    batch_prediction = pipeline.predict(
        batch=[{"text": "test"}, {"text": "no instance for this input"}]
    )
    assert isinstance(batch_prediction, list) and len(batch_prediction) == 2
    assert isinstance(batch_prediction[0], dict) and batch_prediction[1] is None



"""
This module includes all components related to an HPO experiment execution.
It tries to allow for a simple integration with the HPO library 'Ray Tune'.
"""
import logging
import os
import tempfile
from dataclasses import asdict
from datetime import datetime
from pathlib import Path
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import mlflow
from allennlp.data import Vocabulary
from ray import tune
from ray.tune.integration.pytorch_lightning import TuneReportCallback

from biome.text import Pipeline
from biome.text import Trainer
from biome.text import TrainerConfiguration
from biome.text.dataset import Dataset
from biome.text.errors import ValidationError

_LOGGER = logging.getLogger(__name__)


class TuneExperiment(tune.Experiment):
    """This class provides a trainable function and a config to conduct an HPO with `ray.tune.run`

    Parameters
    ----------
    pipeline_config
        The pipeline configuration with its hyperparemter search spaces:
        https://docs.ray.io/en/master/tune/key-concepts.html#search-spaces
    trainer_config
        The trainer configuration with its hyperparameter search spaces
    train_dataset
        Training dataset
    valid_dataset
        Validation dataset
    metrics
        Metrics to report to Tune. If this is a list, each item describes the metric key reported to PyTorch Lightning,
        and it will be reported under the same name to Tune. If this is a dict, each key will be the name reported to
        Tune and the respective value will be the metric key reported to PyTorch Lightning.
        By default (None), all metrics from Pytorch Lightnint will be reported to Tune with the same name.
    vocab
        If you want to share the same vocabulary between the trials you can provide it here
    name
        Used as project name for the WandB logger and for the experiment name in the MLFlow logger.
        By default we construct following string: 'HPO on %date (%time)'
    trainable
        A custom trainable function that takes as input the `TuneExperiment.config` dict.
    mlflow
        If True (default), logs HPO to MLFlow.
    wandb
        If True (default), logs HPO to WandB.
    **kwargs
        The rest of the kwargs are passed on to `tune.Experiment.__init__`.
        They must not contain the 'name', 'run' or the 'config' key,
        since these are provided automatically by `TuneExperiment`.

    Attributes
    ----------
    trainable
        The trainable function used by ray tune
    config
        The config dict passed on to the trainable function

    Examples
    --------
    A minimal usage would be:

    >>> from biome.text import Dataset, TrainerConfiguration
    >>> from ray import tune
    >>> pipeline_config = {
    ...     "name": "tune_experiment_example",
    ...     "head": {"type": "TextClassification", "labels": ["a", "b"]},
    ... }
    >>> trainer_config = TrainerConfiguration(
    ...     optimizer={"type": "adam", "lr": tune.loguniform(1e-3, 1e-2)},
    ...     progress_bar_refresh_rate=0
    ... )
    >>> train_dataset = Dataset.from_dict({"text": ["test", "this"], "label": ["a", "b"]})
    >>> valid_dataset = Dataset.from_dict({"text": ["test", "this"], "label": ["a", "b"]})
    >>> my_exp = TuneExperiment(pipeline_config, trainer_config, train_dataset, valid_dataset, num_samples=10)
    >>> tune.run(my_exp) # doctest: +SKIP

    """

    def __init__(
        self,
        pipeline_config: dict,
        trainer_config: TrainerConfiguration,
        train_dataset: Dataset,
        valid_dataset: Dataset,
        metrics: Union[None, str, List[str], Dict[str, str]] = None,
        vocab: Optional[Vocabulary] = None,
        name: Optional[str] = None,
        trainable: Optional[Callable] = None,
        mlflow: bool = True,
        wandb: bool = True,
        **kwargs,
    ):
        if (
            "name" in kwargs.keys()
            or "run" in kwargs.keys()
            or "config" in kwargs.keys()
        ):
            raise ValueError(
                f"Your `kwargs` must not contain the 'name', 'run' or 'config' key."
                f"These are provided automatically by `TuneExperiment`."
            )

        # save created tmp dirs in this list to clean them up when object gets destroyed
        self._created_tmp_dirs: List[tempfile.TemporaryDirectory] = []

        self._train_dataset_path = self._save_dataset_to_disk(train_dataset)
        self._valid_dataset_path = self._save_dataset_to_disk(valid_dataset)

        self._pipeline_config = pipeline_config
        self._trainer_config = asdict(trainer_config)
        self.trainable = trainable or self._default_trainable

        self._vocab_path = (
            self._save_vocab_to_disk(vocab) if vocab is not None else None
        )
        self._name = name or f"HPO on {datetime.now().strftime('%Y-%m-%d (%I-%M)')}"
        if not os.environ.get("WANDB_PROJECT"):
            os.environ["WANDB_PROJECT"] = self._name

        self._mlflow = mlflow
        self._wandb = wandb
        self._metrics = metrics

        super().__init__(
            name=self._name, run=self.trainable, config=self.config, **kwargs
        )

    def _save_dataset_to_disk(self, dataset: Dataset) -> str:
        """Saves the dataset to disk if not saved already

        Parameters
        ----------
        dataset
            Dataset to save to disk

        Returns
        -------
        dataset_path
            Path to the saved dataset, that is a directory
        """
        try:
            filename = Path(dataset.dataset.cache_files[0]["filename"])
        except (IndexError, KeyError):
            filename = Path()

        if filename.name != "dataset.arrow":
            tmp_dir = tempfile.TemporaryDirectory()
            self._created_tmp_dirs.append(tmp_dir)
            dataset_path = tmp_dir.name
            dataset.save_to_disk(dataset_path)
        else:
            dataset_path = str(filename.absolute())

        # Make sure that we can load the dataset successfully
        try:
            Dataset.load_from_disk(dataset_path)
        except Exception as exception:
            raise ValidationError(
                f"Could not load dataset saved in '{dataset_path}'"
            ) from exception

        return dataset_path

    def _save_vocab_to_disk(self, vocab: Vocabulary) -> str:
        """Saves the vocab to disk to reuse it between the trials

        Parameters
        ----------
        vocab
            Vocabulary to be saved to disk

        Returns
        -------
        vocab_path
            Path to the vocabulary, that is a directory
        """
        tmp_dir = tempfile.TemporaryDirectory()
        self._created_tmp_dirs.append(tmp_dir)
        vocab_path = tmp_dir.name
        vocab.save_to_files(vocab_path)

        # Make sure that we can load the vocab successfully
        try:
            Vocabulary.from_files(vocab_path)
        except Exception as exception:
            raise ValidationError(
                f"Could not load vocab saved in '{vocab_path}'"
            ) from exception

        return vocab_path

    @property
    def config(self) -> dict:
        """The config dictionary used by the `TuneExperiment.trainable` function"""
        return {
            "pipeline_config": self._pipeline_config,
            "trainer_config": self._trainer_config,
            "train_dataset_path": self._train_dataset_path,
            "valid_dataset_path": self._valid_dataset_path,
            "mlflow": self._mlflow,
            "mlflow_tracking_uri": mlflow.get_tracking_uri(),
            "wandb": self._wandb,
            "vocab_path": self._vocab_path,
            "name": self._name,
            "metrics": self._metrics,
        }

    @staticmethod
    def _default_trainable(config, checkpoint_dir=None):
        """A default trainable function used by `tune.run`

        It performs the most straight forward training loop with the provided `config`:
        - Create the pipeline (optionally with a provided vocab)
        - Set up a TuneMetrics logger that reports all metrics back to ray tune after each epoch
        - Execute the training
        """
        pipeline = Pipeline.from_config(
            config["pipeline_config"], vocab_path=config["vocab_path"]
        )

        trainer_config = TrainerConfiguration(**config["trainer_config"])

        callbacks = trainer_config.callbacks
        if not isinstance(callbacks, list):
            callbacks = [callbacks]
        if not any(
            [isinstance(callback, TuneReportCallback) for callback in callbacks]
        ):
            tune_callback = TuneReportCallback(metrics=config["metrics"])
            if trainer_config.callbacks is None:
                trainer_config.callbacks = tune_callback
            else:
                trainer_config.callbacks = callbacks + [tune_callback]

        train_ds = Dataset.load_from_disk(config["train_dataset_path"])
        valid_ds = Dataset.load_from_disk(config["valid_dataset_path"])
        train_instances = train_ds.to_instances(pipeline=pipeline, disable_tqdm=True)
        valid_instances = valid_ds.to_instances(pipeline=pipeline, disable_tqdm=True)

        trainer = Trainer(
            pipeline=pipeline,
            train_dataset=train_instances,
            valid_dataset=valid_instances,
            trainer_config=trainer_config,
            vocab_config=None if config["vocab_path"] else "default",
        )
        trainer.fit()

    def __del__(self):
        """Cleans up the created tmp dirs for the datasets and vocab"""
        for tmp_dir in self._created_tmp_dirs:
            tmp_dir.cleanup()



from typing import Optional

import torch
from allennlp.data import TextFieldTensors
from allennlp.data import Vocabulary
from allennlp.modules import TextFieldEmbedder
from allennlp.modules.seq2seq_encoders import PassThroughEncoder

from .featurizer import InputFeaturizer
from .modules.encoders import Encoder
from .tokenizer import Tokenizer


class ModelBackbone(torch.nn.Module):
    """The backbone of the model.

    It is composed of a tokenizer, featurizer and an encoder.
    This component of the model can be pretrained and used with different task heads.

    Attributes
    ----------
    vocab
        The vocabulary of the pipeline
    featurizer
        Defines the input features of the tokens and indexes
    embedder
        The embedding layer
    encoder
        Outputs an encoded sequence of the tokens
    """

    def __init__(
        self,
        vocab: Vocabulary,
        featurizer: InputFeaturizer,
        embedder: TextFieldEmbedder,
        encoder: Optional[Encoder] = None,
    ):
        super(ModelBackbone, self).__init__()

        self.vocab = vocab
        self.featurizer = featurizer
        self.embedder = embedder
        self.encoder = (
            encoder.input_dim(self.embedder.get_output_dim()).compile()
            if encoder
            else PassThroughEncoder(self.embedder.get_output_dim())
        )

    @property
    def tokenizer(self) -> Tokenizer:
        return self.featurizer.tokenizer

    def forward(
        self, text: TextFieldTensors, mask: torch.Tensor, num_wrapping_dims: int = 0
    ) -> torch.Tensor:
        """Applies the embedding and encoding layer

        Parameters
        ----------
        text
            Output of the `batch.as_tensor_dict()` method, basically the indices of the indexed tokens
        mask
            A mask indicating which one of the tokens are padding tokens
        num_wrapping_dims
            0 if `text` is the output of a `TextField`, 1 if it is the output of a `ListField`

        Returns
        -------
        tensor
            Encoded representation of the input
        """
        embeddings = self.embedder(text, num_wrapping_dims=num_wrapping_dims)
        return self.encoder(embeddings, mask=mask)



from biome.text import helpers
from biome.text.modules.heads.classification.text_classification import (
    TextClassification,
)
from biome.text.modules.heads.classification.text_classification import (
    TextClassificationConfiguration,
)


def test_component_spec_config_with_type():
    head = TextClassificationConfiguration(
        pooler="boe",
        labels=[
            "toxic",
            "severe_toxic",
            "obscene",
            "threat",
            "insult",
            "identity_hate",
        ],
        multilabel=True,
    )

    assert "type" in head.config
    assert head.config["type"] == helpers.get_full_class_name(TextClassification)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import glob

from setuptools import setup

try:
    from setuptools import find_namespace_packages
except ImportError as error:
    raise ImportError("Make sure you have setuptools >= 40.1.0 installed!") from error


if __name__ == "__main__":
    setup(
        name="biome-text",
        use_scm_version=True,
        setup_requires=["setuptools_scm"],
        description="Biome-text is a light-weight open source Natural Language Processing toolbox"
        " built with AllenNLP",
        author="Recognai",
        author_email="francisco@recogn.ai",
        url="https://www.recogn.ai/",
        long_description=open("README.md").read(),
        long_description_content_type="text/markdown",
        packages=find_namespace_packages("src"),
        package_dir={"": "src"},
        install_requires=[
            "allennlp~=2.4.0",
            "beautifulsoup4~=4.9.0",
            "captum~=0.2.0",
            "click~=7.1.0",
            "datasets~=1.2.1",
            "flatdict~=4.0.0",
            "gevent~=20.9.0",
            "ipywidgets~=7.5.1",
            "lxml~=4.6.2",
            "mlflow~=1.13.1",
            "numpy",
            "pandas",
            "pytorch-lightning~=1.3.0",
            "ray[tune]~=1.3.0",
            "spacy~=3.0.0",
            "torch",  # the version is defined by allennlp
            "transformers",  # the version is defined by allennlp
            "tqdm>=4.49.0",
            "fastapi~=0.63.0",  # newer versions brings pydantic conflicts with spaCy 3.0.x
            "uvicorn>=0.13.0",
            "pyyaml",
        ],
        extras_require={
            "dev": [
                # testing
                "pytest>=6.2.0",
                "pytest-cov>=2.10.0",
                "pytest-pylint>=0.14.0",
                "pytest-notebook~=0.6.0",
                "wandb>=0.10.12",
                "xlrd~=1.2.0",
                # documentation
                "pdoc3~=0.8.1",
                # development
                "pre-commit~=2.9.0",
                "GitPython",
            ],
        },
        entry_points={"console_scripts": ["biome=biome.text.cli:main"]},
        python_requires=">=3.6.1",  # taken from AllenNLP
        zip_safe=False,
    )



import logging
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union
from typing import cast

import torch
from allennlp.data import Instance
from allennlp.data import TextFieldTensors
from allennlp.data import Token
from allennlp.data.fields import MetadataField
from allennlp.data.fields import SequenceLabelField
from allennlp.data.fields import TextField
from allennlp.modules import ConditionalRandomField
from allennlp.modules import FeedForward
from allennlp.modules import TimeDistributed
from allennlp.modules.conditional_random_field import allowed_transitions
from allennlp.nn.util import get_text_field_mask
from spacy.tokens.doc import Doc
from spacy.vocab import Vocab

from biome.text import vocabulary
from biome.text.backbone import ModelBackbone
from biome.text.errors import WrongValueError
from biome.text.featurizer import FeaturizeError
from biome.text.helpers import offsets_from_tags
from biome.text.helpers import spacy_to_allennlp_token
from biome.text.helpers import span_labels_to_tag_labels
from biome.text.helpers import tags_from_offsets
from biome.text.metrics import Metrics
from biome.text.modules.configuration import ComponentConfiguration
from biome.text.modules.configuration import FeedForwardConfiguration
from biome.text.modules.heads.task_head import TaskHead
from biome.text.modules.heads.task_head import TaskName
from biome.text.modules.heads.task_prediction import Entity
from biome.text.modules.heads.task_prediction import TokenClassificationPrediction


class TokenClassification(TaskHead):
    """Task head for token classification (NER)

    Parameters
    ----------
    backbone
        The model backbone
    labels
        List span labels. Span labels get converted to tag labels internally, using
        configured label_encoding for that.
    label_encoding
        The format of the tags. Supported encodings are: ['BIO', 'BIOUL']
    top_k
    dropout
    feedforward
    """

    _LOGGER = logging.getLogger(__name__)

    task_name = TaskName.token_classification

    def __init__(
        self,
        backbone: ModelBackbone,
        labels: List[str],
        label_encoding: Optional[str] = "BIOUL",
        top_k: int = 1,
        dropout: Optional[float] = 0.0,
        feedforward: Optional[FeedForwardConfiguration] = None,
    ) -> None:
        super().__init__(backbone)

        if label_encoding not in ["BIOUL", "BIO"]:
            raise WrongValueError(
                f"Label encoding {label_encoding} not supported. Allowed values are {['BIOUL', 'BIO']}"
            )

        self._span_labels = labels
        self._label_encoding = label_encoding

        vocabulary.set_labels(
            self.backbone.vocab,
            # Convert span labels to tag labels if necessary
            # We just check if "O" is in the label list, a necessary tag for IOB/BIOUL schemes,
            # an unlikely label for spans
            span_labels_to_tag_labels(labels, self._label_encoding),
        )

        self.top_k = top_k
        self.dropout = torch.nn.Dropout(dropout)
        self._feedforward: FeedForward = (
            None
            if not feedforward
            else feedforward.input_dim(backbone.encoder.get_output_dim()).compile()
        )
        # output layers
        self._classifier_input_dim = (
            self._feedforward.get_output_dim()
            if self._feedforward
            else backbone.encoder.get_output_dim()
        )
        # we want this linear applied to each token in the sequence
        self._label_projection_layer = TimeDistributed(
            torch.nn.Linear(self._classifier_input_dim, self.num_labels)
        )
        constraints = allowed_transitions(
            self._label_encoding,
            vocabulary.get_index_to_labels_dictionary(self.backbone.vocab),
        )

        self._crf = ConditionalRandomField(
            self.num_labels, constraints, include_start_end_transitions=True
        )

        # There is no top_k option for the f1 metric, it will always only take into account the first choice
        # If you want to use top_k in the accuracy, you have to change the way we convert the CRF output to logits!
        self._metrics = Metrics(
            accuracy={"type": "categorical_accuracy"},
            f1={
                "type": "span_f1",
                "vocabulary": self.backbone.vocab,
                "tag_namespace": vocabulary.LABELS_NAMESPACE,
                "label_encoding": self._label_encoding,
            },
        )

    @property
    def span_labels(self) -> List[str]:
        return self._span_labels

    def _loss(self, logits: torch.Tensor, labels: torch.Tensor, mask: torch.Tensor):
        """loss is calculated as -log_likelihood from crf"""
        return -1 * self._crf(logits, labels, mask)

    def featurize(
        self,
        text: Union[str, List[str]],
        entities: Optional[List[dict]] = None,
        tags: Optional[Union[List[str], List[int]]] = None,
    ) -> Instance:
        """
        Parameters
        ----------
        text
            Can be either a simple str or a list of str,
            in which case it will be treated as a list of pretokenized tokens
        entities
            A list of span labels

            Span labels are dictionaries that contain:

            'start': int, char index of the start of the span
            'end': int, char index of the end of the span (exclusive)
            'label': str, label of the span

            They are used with the `spacy.gold.biluo_tags_from_offsets` method.
        tags
            A list of tags in the BIOUL or BIO format.
        """
        if isinstance(text, str):
            doc = self.backbone.tokenizer.nlp(text)
            tokens = [spacy_to_allennlp_token(token) for token in doc]
            tags = (
                tags_from_offsets(doc, entities, self._label_encoding)
                if entities is not None
                else []
            )
            # discard misaligned examples for now
            if "-" in tags:
                raise FeaturizeError(
                    f"Could not align spans with tokens for following example: '{text}' {entities}"
                )
        # text is already pre-tokenized
        else:
            tokens = [Token(t) for t in text]

        instance = self.backbone.featurizer(
            tokens, to_field="text", tokenize=False, aggregate=True
        )

        if self.training:
            try:
                instance.add_field(
                    "tags",
                    SequenceLabelField(
                        tags,
                        sequence_field=cast(TextField, instance["text"]),
                        label_namespace=vocabulary.LABELS_NAMESPACE,
                    ),
                )
            except Exception as exception:
                raise FeaturizeError(
                    f"Could not create SequenceLabelField for {(tokens, tags)}"
                ) from exception

        instance.add_field("raw_text", MetadataField(text))

        return instance

    def forward(  # type: ignore
        self,
        text: TextFieldTensors,
        raw_text: List[Union[str, List[str]]],
        tags: torch.IntTensor = None,
    ) -> Dict:

        mask = get_text_field_mask(text)
        embedded_text = self.dropout(self.backbone.forward(text, mask))

        if self._feedforward is not None:
            embedded_text = self._feedforward(embedded_text)

        logits = self._label_projection_layer(embedded_text)
        # `self._crf.viterbi_tags` can return invalid tag sequences when logits are nan
        viterbi_logits = torch.where(
            torch.isnan(logits), torch.zeros_like(logits), logits
        )
        # dims are: batch, top_k, (tag_sequence, viterbi_score)
        viterbi_paths: List[List[Tuple[List[int], float]]] = self._crf.viterbi_tags(
            viterbi_logits, mask, top_k=self.top_k
        )
        # We just keep the best path for every instance
        predicted_tags: List[List[int]] = [paths[0][0] for paths in viterbi_paths]
        class_probabilities = torch.zeros_like(logits)

        for i, instance_tags in enumerate(predicted_tags):
            for j, tag_id in enumerate(instance_tags):
                class_probabilities[i, j, tag_id] = 1

        output = dict(
            viterbi_paths=viterbi_paths,
            raw_text=raw_text,
        )

        if tags is not None:
            output["loss"] = self._loss(logits, tags, mask)
            for metric in self._metrics.get_dict(is_train=self.training).values():
                metric(class_probabilities, tags, mask)

        return output

    def _make_task_prediction(
        self,
        single_forward_output: Dict,
        instance: Instance,
    ) -> TokenClassificationPrediction:
        # The dims are: top_k, tags
        tags: List[List[str]] = self._make_tags(single_forward_output["viterbi_paths"])
        # construct a spacy Doc
        pre_tokenized = not isinstance(single_forward_output["raw_text"], str)
        if pre_tokenized:
            # compose doc from tokens
            doc = Doc(Vocab(), words=single_forward_output["raw_text"])
        else:
            doc = self.backbone.tokenizer.nlp(single_forward_output["raw_text"])

        return TokenClassificationPrediction(
            tags=tags,
            scores=[score for tags, score in single_forward_output["viterbi_paths"]],
            entities=self._make_entities(doc, tags, pre_tokenized),
        )

    def _make_tags(
        self, viterbi_paths: List[Tuple[List[int], float]]
    ) -> List[List[str]]:
        """Makes the 'tags' key of the task prediction"""
        return [
            [vocabulary.label_for_index(self.backbone.vocab, idx) for idx in tags]
            for tags, score in viterbi_paths
        ]

    def _make_entities(
        self,
        doc: Doc,
        k_tags: List[List[str]],
        pre_tokenized: bool,
    ) -> List[List[Entity]]:
        """Makes the 'entities' key of the task prediction. Computes offsets with respect to char and token id"""
        return [
            [
                Entity(**entity)
                for entity in offsets_from_tags(
                    doc, tags, self._label_encoding, only_token_spans=pre_tokenized
                )
            ]
            for tags in k_tags
        ]

    def get_metrics(self, reset: bool = False) -> Dict[str, float]:
        metrics: Dict[str, float] = {}
        for name, metric in self._metrics.get_dict(is_train=self.training).items():
            metric_value = metric.get_metric(reset=reset)
            try:
                metrics.update(metric_value)
            # AllenNLP's CategoricalAccuracy does not comply with AllenNLP's Metric API
            except TypeError:
                metrics[name] = metric_value

        return metrics


class TokenClassificationConfiguration(ComponentConfiguration[TokenClassification]):
    """Configuration for classification head components"""

    pass



import os
from pathlib import Path
from typing import Optional

import click

from biome.text import Dataset
from biome.text import Pipeline
from biome.text import Trainer
from biome.text import TrainerConfiguration
from biome.text.helpers import yaml_to_dict


@click.command()
@click.argument(
    "pipeline_path",
    type=click.Path(exists=True),
    required=True,
)
@click.option(
    "--output",
    "-o",
    type=click.Path(),
    required=True,
    help="Path of the training output.",
)
@click.option(
    "--trainer_config",
    type=click.Path(exists=True),
    required=True,
    help="Path to the trainer configuration YAML file.",
)
@click.option(
    "--train_data",
    type=click.Path(exists=True),
    required=True,
    help="Path to the training data.",
)
@click.option(
    "--valid_data",
    type=click.Path(exists=True),
    required=False,
    help="Path to the validation data.",
)
def train(
    pipeline_path: str,
    output: str,
    trainer_config: str,
    train_data: str,
    valid_data: Optional[str] = None,
) -> None:
    """Train a pipeline.

    PIPELINE_PATH is either the path to a pretrained pipeline (model.tar.gz file),
    or the path to a pipeline configuration (YAML file).
    """
    _, extension = os.path.splitext(pipeline_path)
    extension = extension[1:].lower()
    pipeline = (
        Pipeline.from_yaml(pipeline_path)
        if extension in ["yaml", "yml"]
        else Pipeline.from_pretrained(pipeline_path)
    )

    datasets = {
        "train": dataset_from_path(train_data),
        "validation": dataset_from_path(valid_data) if valid_data else None,
    }

    trainer = Trainer(
        pipeline=pipeline,
        train_dataset=datasets["train"],
        valid_dataset=datasets["validation"],
        trainer_config=TrainerConfiguration(**yaml_to_dict(trainer_config)),
    )
    trainer.fit(output_dir=output)


def dataset_from_path(path: str) -> Dataset:
    file_extension = Path(path).suffix
    if file_extension in [".csv"]:
        return Dataset.from_csv(path)
    elif file_extension in [".json", ".jsonl"]:
        return Dataset.from_json(path)
    else:
        raise ValueError(
            f"Could not create a Dataset from '{path}'. "
            f"We only support following formats: [csv, json, jsonl]"
        )



from typing import Optional

import click

from biome.text import Pipeline
from biome.text.cli.train import dataset_from_path


@click.command()
@click.argument(
    "pipeline_path",
    type=click.Path(exists=True),
)
@click.option(
    "--output",
    "-o",
    type=click.Path(),
    required=True,
    help="Path to write the evaluation metrics to.",
)
@click.option(
    "--dataset",
    "-ds",
    type=click.Path(exists=True),
    required=True,
    help="Path to the dataset",
)
@click.option(
    "--batch_size",
    "-bs",
    type=int,
    default=16,
    show_default=True,
    help="Batch size during evaluation.",
)
@click.option(
    "--lazy",
    "-l",
    type=bool,
    default=False,
    show_default=True,
    help="If true, data is lazily loaded from disk, otherwise it is loaded into memory.",
)
@click.option(
    "--prediction_output",
    "-po",
    type=click.Path(),
    default=None,
    help="Write batch predictions to this file.",
)
def evaluate(
    pipeline_path: str,
    output: str,
    dataset: str,
    batch_size: int = 16,
    lazy: bool = False,
    prediction_output: Optional[str] = None,
) -> None:
    """Evaluate a pipeline on a given dataset.

    PIPELINE_PATH is the path to a pretrained pipeline (model.tar.gz file).
    """
    pipeline = Pipeline.from_pretrained(pipeline_path)
    dataset = dataset_from_path(dataset)

    pipeline.evaluate(
        dataset,
        batch_size=batch_size,
        lazy=lazy,
        predictions_output_file=prediction_output,
        metrics_output_file=output,
    )



import logging
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union
from typing import cast

import numpy
import torch
from allennlp.data import Instance
from allennlp.data import TextFieldTensors
from allennlp.data.fields import TextField
from allennlp.modules.seq2vec_encoders import BagOfEmbeddingsEncoder
from allennlp.nn.util import get_text_field_mask
from captum.attr import IntegratedGradients

from biome.text.backbone import ModelBackbone
from biome.text.modules.configuration import ComponentConfiguration
from biome.text.modules.configuration import FeedForwardConfiguration
from biome.text.modules.configuration import Seq2VecEncoderConfiguration
from biome.text.modules.heads.classification.classification import ClassificationHead
from biome.text.modules.heads.task_prediction import Attribution
from biome.text.modules.heads.task_prediction import TextClassificationPrediction


class TextClassification(ClassificationHead):
    """
    Task head for text classification

    Parameters
    ----------
    backbone
        The backbone of your model. Must not be provided when initiating with `Pipeline.from_config`.
    labels
        A list of labels for your classification task.
    pooler
        The pooler of the output sequence from the backbone model. Default: `BagOfEmbeddingsEncoder`.
    feedforward
        An optional feedforward layer applied to the output of the pooler. Default: None.
    multilabel
        Is this a multi label classification task? Default: False
    label_weights
        A list of weights for each label. The weights must be in the same order as the `labels`.
        You can also provide a dictionary that maps the label to its weight. Default: None.
    """

    label_name = "label"
    forward_arg_name = "text"
    _LOGGER = logging.getLogger(__name__)

    def __init__(
        self,
        backbone: ModelBackbone,
        labels: List[str],
        pooler: Optional[Seq2VecEncoderConfiguration] = None,
        feedforward: Optional[FeedForwardConfiguration] = None,
        multilabel: bool = False,
        label_weights: Optional[Union[List[float], Dict[str, float]]] = None,
    ) -> None:

        super().__init__(backbone, labels, multilabel, label_weights=label_weights)

        self.pooler = (
            pooler.input_dim(self.backbone.encoder.get_output_dim()).compile()
            if pooler
            else BagOfEmbeddingsEncoder(self.backbone.encoder.get_output_dim())
        )
        self.feedforward = (
            None
            if not feedforward
            else feedforward.input_dim(self.pooler.get_output_dim()).compile()
        )
        self._classification_layer = torch.nn.Linear(
            (self.feedforward or self.pooler).get_output_dim(), self.num_labels
        )

    def featurize(
        self,
        text: Union[str, List[str], Dict[str, str]],
        label: Optional[Union[str, List[str]]] = None,
    ) -> Instance:
        instance = self.backbone.featurizer(
            text,
            to_field=self.forward_arg_name,
            aggregate=True,
            exclude_record_keys=True,
        )

        return self._add_label(instance, label, to_field=self.label_name)

    def forward(  # type: ignore
        self,
        text: TextFieldTensors,
        label: torch.IntTensor = None,
    ) -> Dict[str, Any]:

        mask = get_text_field_mask(text)
        embeddings = self.backbone.embedder(text)
        logits = self._encoder_and_head_forward(embeddings, mask)

        output = self._make_forward_output(logits, label)

        # For computing the attributions
        # TODO: An optimized implementation would be to calculate the attributions directly in the forward method
        #  and provide a practical switch, maybe: `with head.turn_attributions_on(): self.forward_on_instances()`
        #  In this way we would calculate the attributions batch wise and on on GPU if available.
        if not self.training:
            output["embeddings"], output["mask"] = embeddings, mask

        return output

    def _encoder_and_head_forward(
        self, embeddings: torch.Tensor, mask: torch.Tensor
    ) -> torch.Tensor:
        """We reuse this method for the computation of the attributions"""
        encoded_text = self.pooler(
            self.backbone.encoder(embeddings, mask=mask), mask=mask
        )
        if self.feedforward:
            encoded_text = self.feedforward(encoded_text)

        return self._classification_layer(encoded_text)

    def _compute_attributions(
        self,
        single_forward_output,
        instance,
        n_steps: int = 50,
    ) -> List[Attribution]:
        """Attributes the prediction to the input.

        The attributions are calculated by means of the [Integrated Gradients](https://arxiv.org/abs/1703.01365) method.

        Parameters
        ----------
        single_forward_output
            Non-batched forward output containing numpy arrays
        instance
            The instance containing the input data
        n_steps
            The number of steps used when calculating the attribution of each token.

        Returns
        -------
        attributions
            A list of attributions
        """
        # captum needs `torch.Tensor`s and we need a batch dimension (-> unsqueeze)
        embeddings = torch.from_numpy(single_forward_output["embeddings"]).unsqueeze(0)
        logits = torch.from_numpy(single_forward_output["logits"]).unsqueeze(0)
        mask = torch.from_numpy(single_forward_output["mask"]).unsqueeze(0)

        ig = IntegratedGradients(self._encoder_and_head_forward)
        attributions, delta = ig.attribute(
            embeddings,
            n_steps=n_steps,
            target=torch.argmax(logits),
            additional_forward_args=mask,
            return_convergence_delta=True,
        )
        attributions = attributions.sum(dim=2).squeeze(0)
        attributions = attributions / torch.norm(attributions)
        attributions = attributions.detach().numpy()

        text_tokens = cast(TextField, instance[self.forward_arg_name]).tokens

        return [
            Attribution(
                text=token.text,
                start=token.idx,
                end=self._get_token_end(token),
                field=self.forward_arg_name,
                attribution=attribution,
            )
            for token, attribution in zip(text_tokens, attributions)
        ]

    def _make_task_prediction(
        self,
        single_forward_output: Dict[str, numpy.ndarray],
        instance: Instance,
    ) -> TextClassificationPrediction:
        labels, probabilities = self._compute_labels_and_probabilities(
            single_forward_output
        )

        return TextClassificationPrediction(labels=labels, probabilities=probabilities)


class TextClassificationConfiguration(ComponentConfiguration[TextClassification]):
    """Configuration for classification head components"""

    pass



from allennlp.data import Vocabulary

from biome.text.metrics import Metrics


def test_metrics():
    metrics = Metrics(
        accuracy={"type": "categorical_accuracy"},
        f1={
            "type": "span_f1",
            "vocabulary": Vocabulary.empty(),
        },
    )

    # Check that training and validation metrics are different instances
    assert (
        metrics.get_dict()["accuracy"]
        is not metrics.get_dict(is_train=False)["accuracy"]
    )
    # Check if we share the same vocab
    assert (
        metrics.get_dict()["f1"]._label_vocabulary
        is metrics.get_dict(is_train=False)["f1"]._label_vocabulary
    )



import os
from pathlib import Path
from tempfile import mkdtemp

import pytest

from biome.text import Dataset
from biome.text import Pipeline
from biome.text import PipelineConfiguration
from biome.text.configuration import FeaturesConfiguration
from biome.text.configuration import VocabularyConfiguration
from biome.text.modules.heads import TaskHeadConfiguration
from biome.text.modules.heads import TextClassification


class MyCustomHead(TextClassification):
    """Just a head renaming the original TextClassification head"""

    pass


@pytest.fixture
def training_dataset() -> Dataset:
    """Creates the training dataset and gives the structure"""
    resources_path = (
        Path(__file__).parent.parent.parent / "tests" / "resources" / "data"
    )
    training_ds = Dataset.from_csv(paths=str(resources_path / "dataset_source.csv"))

    # Keeping just 'label' and text 'category'
    training_ds = training_ds.map(
        lambda x: {"label": x["job"], "text": x["education"] + " " + x["marital"]},
    )

    return training_ds


def test_load_pipeline_with_custom_head(training_dataset, tmp_path):
    """Testing a model training inserting a class as custom heard"""

    # Pipeline configuration dict with custom head
    config = PipelineConfiguration(
        "test-pipeline",
        head=TaskHeadConfiguration(
            type=MyCustomHead,
            labels=[
                "blue-collar",
                "technician",
                "management",
                "services",
                "retired",
                "admin.",
            ],
        ),
        features=FeaturesConfiguration(),
    )

    # Asserting that pipeline.head is an instance of MyCustomHead
    pipeline = Pipeline.from_config(config)
    assert isinstance(pipeline.head, MyCustomHead)

    # Saving the pipeline to output
    output = tmp_path / "pipeline"
    pipeline.save(output)

    # Loading model from output
    trained_pl = Pipeline.from_pretrained(os.path.join(str(output), "model.tar.gz"))
    trained_pl.predict("Oh yeah")

    # Asserting that the pipeline head is recognized as `MyCustomHead` instance after loading from a model.tar.gz
    assert isinstance(trained_pl.head, MyCustomHead)



import re

import pytest
from pytest_notebook.nb_regression import NBRegressionFixture
from pytest_notebook.notebook import dump_notebook
from pytest_notebook.notebook import load_notebook

pytestmark = pytest.mark.skip(
    reason="The pytest-notebook package is not actively maintained and "
    "the tutorial tests are quite heavy on resources. "
    "The idea is to run those tests locally and manually from time to time."
    "THESE TESTS ARE ALSO OUT OF DATE ... :/"
)


def test_text_classifier_tutorial(tmp_path, tutorials_path):
    notebook_path = tutorials_path / "Training_a_text_classifier.ipynb"

    # adapt notebook to CI (make its execution quicker + comment lines)
    notebook = load_notebook(str(notebook_path))
    for cell in notebook["cells"]:
        if cell["source"].startswith("!pip install"):
            cell["source"] = re.sub(r"!pip install", r"#!pip install", cell["source"])
        if cell["source"].startswith("trainer_config ="):
            cell["source"] = re.sub(
                r"num_epochs=[0-9][0-9]?", r"num_epochs=1", cell["source"]
            )
        if cell["source"].startswith("pl.train("):
            cell["source"] = re.sub(
                r"training=train_ds", r"training=valid_ds", cell["source"]
            )
    # dump adapted notebook
    mod_notebook_path = tmp_path / notebook_path.name
    with mod_notebook_path.open("w") as file:
        file.write(str(dump_notebook(notebook)))

    # test adapted notebook
    fixture = NBRegressionFixture(exec_timeout=100)
    fixture.check(str(mod_notebook_path))


def test_slot_filling_tutorial(tmp_path, tutorials_path):
    notebook_path = tutorials_path / "Training_a_sequence_tagger_for_Slot_Filling.ipynb"

    # adapt notebook to CI (make its execution quicker + comment lines)
    notebook = load_notebook(str(notebook_path))
    for cell in notebook["cells"]:
        if cell["source"].startswith("!pip install"):
            cell["source"] = re.sub(r"!pip install", r"#!pip install", cell["source"])
        if cell["source"].startswith(
            "from biome.text.configuration import FeaturesConfiguration"
        ):
            cell["source"] = re.sub(
                r"https://dl.fbaipublicfiles.com/fasttext/vectors-english/wiki-news-300d-1M.vec.zip",
                r"https://biome-tutorials-data.s3-eu-west-1.amazonaws.com/token_classifier/wiki-news-300d-1M.head.vec",
                cell["source"],
            )
        if cell["source"].startswith("trainer_config ="):
            cell["source"] = re.sub(
                r"TrainerConfiguration\(\)",
                r"TrainerConfiguration(num_epochs=1)",
                cell["source"],
            )
        if cell["source"].startswith("pl.train("):
            cell["source"] = re.sub(
                r"pl.train",
                r"from biome.text.configuration import TrainerConfiguration\npl.train",
                cell["source"],
            )
            cell["source"] = re.sub(
                r"training=train_ds",
                r"training=valid_ds",
                cell["source"],
            )
            cell["source"] = re.sub(
                r"test=test_ds,",
                r"test=test_ds, trainer=TrainerConfiguration(num_epochs=1)",
                cell["source"],
            )

    # dump adapted notebook
    mod_notebook_path = tmp_path / notebook_path.name
    with mod_notebook_path.open("w") as file:
        file.write(str(dump_notebook(notebook)))

    # test adapted notebook
    fixture = NBRegressionFixture(exec_timeout=200)
    fixture.check(str(mod_notebook_path))



import pytest
from numpy.testing import assert_allclose

from biome.text import Pipeline
from biome.text.modules.heads.task_prediction import Attribution
from biome.text.modules.heads.task_prediction import DocumentClassificationPrediction


@pytest.fixture
def pipeline() -> Pipeline:
    labels = ["a", "b", "c", "d", "e", "f"]
    return Pipeline.from_config(
        {
            "name": "test_document_classification",
            "tokenizer": {"segment_sentences": False},
            "head": {"type": "DocumentClassification", "labels": labels},
        }
    )


@pytest.mark.parametrize(
    "segment_sentences, input, output",
    [
        (False, "one sentence. two sentence", (1, 5)),
        (True, "one sentence. two sentence", (2, 3)),
        (False, ["one sentence. two sentence", "test"], (2, 5)),
        (True, ["one sentence. two sentence", "test"], (3, 3)),
        (False, {"one": "one sentence. two sentence", "two": "test"}, (2, 5)),
        (True, {"one": "one sentence. two sentence", "two": "test"}, (3, 3)),
    ],
)
def test_tokenization_of_different_input(pipeline, segment_sentences, input, output):
    pipeline = Pipeline.from_config(
        {
            "name": "test_document_classification",
            "tokenizer": {"segment_sentences": segment_sentences},
            "head": {"type": "DocumentClassification", "labels": "a"},
        }
    )
    instance = pipeline.head.featurize(input)
    tokens = pipeline.head._extract_tokens(instance)

    assert len(tokens) == output[0]
    assert len(tokens[0]) == output[1]


def test_make_task_prediction(pipeline):
    instance = pipeline.head.featurize("test this sentence")
    forward_output = pipeline._model.forward_on_instances([instance])

    prediction = pipeline.head._make_task_prediction(forward_output[0], None)

    assert isinstance(prediction, DocumentClassificationPrediction)
    assert isinstance(prediction.labels, list) and isinstance(
        prediction.probabilities, list
    )
    assert len(prediction.labels) == len(prediction.probabilities) == 6
    # check descending order
    assert_allclose(
        sorted(prediction.probabilities, reverse=True), prediction.probabilities
    )
    assert all([isinstance(label, str) for label in prediction.labels])
    assert set(pipeline.head.labels) == set(prediction.labels)
    assert all([isinstance(prob, float) for prob in prediction.probabilities])


def test_compute_attributions(pipeline):
    instance = pipeline.head.featurize("test this sentence")
    pipeline.model.eval()
    forward_output = pipeline.model.forward_on_instances([instance])

    attributions = pipeline.head._compute_attributions(
        forward_output[0], instance, n_steps=1
    )

    assert isinstance(attributions, list) and isinstance(attributions[0], list)
    assert len(attributions) == 1 and len(attributions[0]) == 3
    assert all(
        [isinstance(attribution, Attribution) for attribution in attributions[0]]
    )
    assert all([attr.field == "text" for attr in attributions[0]])
    assert all([isinstance(attr.attribution, float) for attr in attributions[0]])
    assert attributions[0][1].start == 5 and attributions[0][1].end == 9
